#define export __declspec (dllexport)
#include <windows.h>
#include <stdio.h>
#include <process.h>
#define maxopen 5

typedef struct sNode {
	long fs;
	long ts;
	FILE* ff;
	FILE* tf;
	int done;
} tNode;
tNode cp[maxopen];
CRITICAL_SECTION* cs;
long processor;
char rs[16];

export char* getMod(char *f)
{
	HANDLE h;
	FILETIME ft;
	SYSTEMTIME st;
	h = CreateFile(f,GENERIC_READ,FILE_SHARE_READ,
		NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (h == INVALID_HANDLE_VALUE) return "";
	if (!GetFileTime(h,NULL,NULL,&ft))
	{
		CloseHandle(h);
		return "";
	}
	CloseHandle(h);
	FileTimeToSystemTime(&ft, &st);
	sprintf(rs,"%u%02u%02u%02u%02u%02u",st.wYear,st.wMonth,st.wDay,st.wHour,st.wMinute,st.wSecond);
	return rs;
}

export double setMod(char *f, char *t)
{
	HANDLE h;
	FILETIME ft;
	SYSTEMTIME st;
	int r1,r2,r3,r4,r5,r6;
	h = CreateFile(f,GENERIC_WRITE,FILE_SHARE_WRITE,
		NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if (h == INVALID_HANDLE_VALUE) return 1;
	sscanf(t,"%4i%2i%2i%2i%2i%2i",&r1,&r2,&r3,&r4,&r5,&r6);
	st.wYear = r1;
	st.wMonth = r2;
	st.wDay = r3;
	st.wHour = r4;
	st.wMinute = r5;
	st.wSecond = r6;
	SystemTimeToFileTime(&st,&ft);
	if (!SetFileTime(h,NULL,NULL,&ft))
	{
		CloseHandle(h);
		return 2;
	}
	CloseHandle(h);
	return 0;
}

export double init()
{
	cs = malloc(sizeof(CRITICAL_SECTION));
	InitializeCriticalSection(cs);
	return 0;
}

export double kill()
{
	int n;
	for (n = 0; n < maxopen; ++n)
	{
		EnterCriticalSection(cs);
		if (cp[n].done == 1)
		{
			fclose(cp[n].ff);
			fclose(cp[n].tf);
		}
		LeaveCriticalSection(cs);
	}
	DeleteCriticalSection(cs);
	free(cs);
	cs = NULL;
	return 0;
}

unsigned WINAPI proc_thread(void* arg)
{
	unsigned char fb[BUFSIZ];
	size_t bs;
	int n;
	int finished = 0;
	while (finished == 0)
	{
		for (n = 0; n < maxopen; ++n)
		{
			EnterCriticalSection(cs);
				if (cp[n].done == 1)
				{
					if ((bs = fread(fb, 1, BUFSIZ, cp[n].ff)) == 0)
					{
						fclose(cp[n].ff);
						fclose(cp[n].tf);
						cp[n].done = 0;
					}
					else
					{
						fwrite(fb, 1, bs, cp[n].tf);
						cp[n].ts += bs;
					}
				}
				else
					finished++;
			LeaveCriticalSection(cs);
		}
		if (finished < maxopen)
			finished = 0;
	}
	CloseHandle((HANDLE)processor);
	processor = 0;
	_endthreadex(0);
	return 0;
}

export double copy(char *from, char *to)
{
	int n;
	for (n = 0; n < maxopen + 1; ++n)
	{
		if (n == maxopen) return -1;
		EnterCriticalSection(cs);
		if (cp[n].done == 0) break;
		LeaveCriticalSection(cs);
	}
	//open files and prepare node
	//EnterCriticalSection(cs);
	cp[n].ts = 0;
	cp[n].ff = fopen(from,"rb");
	if (cp[n].ff == NULL)
	{
		LeaveCriticalSection(cs);
		return -2;
	}
	fseek(cp[n].ff,0L,SEEK_END);
	cp[n].fs = ftell(cp[n].ff);
	fseek(cp[n].ff,0L,SEEK_SET);
	cp[n].tf = fopen(to,"wb");
	if (cp[n].tf == NULL)
	{
		fclose(cp[n].ff);
		LeaveCriticalSection(cs);
		return -3;
	}
	cp[n].done = 1;
	LeaveCriticalSection(cs);
	//singleton processor - create only if one doesn't exist
	if (processor == 0)
	{
		processor = _beginthreadex(NULL, 0, &proc_thread, NULL, 0, NULL);
		if (processor == 0)
		{
			EnterCriticalSection(cs);
			cp[n].done = 0;
			fclose(cp[n].ff);
			fclose(cp[n].tf);
			LeaveCriticalSection(cs);
			return -4;
		}
	}
	return (double)n;
}

export double cancel(double n)
{
	EnterCriticalSection(cs);
		cp[(int)n].done = 0;
		fclose(cp[(int)n].ff);
		fclose(cp[(int)n].tf);
	LeaveCriticalSection(cs);
	return 0;
}

export double isize(double n)
{
	double r;
	EnterCriticalSection(cs);
		r = cp[(int)n].fs;
	LeaveCriticalSection(cs);
	return r;
}

export double osize(double n)
{
	double r;
	EnterCriticalSection(cs);
		r = cp[(int)n].ts;
	LeaveCriticalSection(cs);
	return r;
}

export double progress(double n)
{
	double r = -1;
	EnterCriticalSection(cs);
		if (cp[(int)n].done == 0)
			r = 2;
		else if (cp[(int)n].fs != 0L)
			r = (double)cp[(int)n].ts / (double)cp[(int)n].fs;
	LeaveCriticalSection(cs);
	return r;
}
